#!/usr/bin/env python
# shutdowner.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (shutdowner.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
#
"""
.. module:: shutdowner.

.. raw:: html

    <a href="https://bitdust.io/automats/shutdowner/shutdowner.png" target="_blank">
    <img src="https://bitdust.io/automats/shutdowner/shutdowner.png" style="max-width:100%;">
    </a>

The state machine ``shutdowner()`` manages the completion of the program.

Synchronized between the completion of the blocking code and stop of the main Twisted reactor.

State "FINISHED" is a sign for the ``initializer()`` automat to stop of the whole program.


EVENTS:
    * :red:`block`
    * :red:`init`
    * :red:`reactor-stopped`
    * :red:`ready`
    * :red:`stop`
    * :red:`unblock`
"""

from __future__ import absolute_import
import six

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

import sys

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in shutdowner.py')

from twisted.internet.defer import DeferredList

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.main import initializer
from bitdust.main import settings
from bitdust.main import config

#------------------------------------------------------------------------------

_Shutdowner = None

#------------------------------------------------------------------------------


def shutdown_settings():
    if config.conf():
        config.conf().removeConfigNotifier('logs/debug-level')
    settings.shutdown()


def shutdown_engine():
    from bitdust.contacts import identitydb
    from bitdust.userid import id_url
    from bitdust.main import listeners
    from bitdust.main import events
    identitydb.shutdown()
    id_url.shutdown()
    listeners.shutdown()
    events.shutdown()


def shutdown_automats():
    survived_automats = list(automat.objects().values())
    if survived_automats:
        lg.warn('found %d survived state machines, sending "shutdown" event to them all' % len(survived_automats))
        for a in survived_automats:
            if a.name != 'shutdowner':
                a.event('shutdown')
    survived_automats = list(automat.objects().values())
    if survived_automats:
        lg.warn('still found %d survived state machines, executing "destroy()" method to them all' % len(survived_automats))
        for a in survived_automats:
            if a.name != 'shutdowner':
                a.destroy()
    automat.shutdown()


def shutdown_local():
    from bitdust.logs import weblog
    from bitdust.logs import webtraffic
    from bitdust.system import tmpfile
    from bitdust.system import run_upnpc
    from bitdust.raid import eccmap
    from bitdust.lib import net_misc
    from bitdust.updates import git_proc
    from bitdust.crypt import my_keys
    from bitdust.userid import my_id
    my_keys.shutdown()
    my_id.shutdown()
    eccmap.shutdown()
    run_upnpc.shutdown()
    net_misc.shutdown()
    git_proc.shutdown()
    tmpfile.shutdown()
    try:
        weblog.shutdown()
    except:
        lg.exc()
    try:
        webtraffic.shutdown()
    except:
        lg.exc()


def shutdown_interfaces():
    from bitdust.interface import api_rest_http_server
    from bitdust.interface import api_web_socket
    # from bitdust.interface import ftp_server
    # ftp_server.shutdown()
    api_rest_http_server.shutdown()
    api_web_socket.shutdown()


def shutdown_services():
    from bitdust.services import driver
    driver.shutdown()


def shutdown(x=None):
    """
    This is a top level method which control the process of finishing the
    program.

    Calls method ``shutdown()`` in other modules.
    """
    if _Debug:
        lg.out(_DebugLevel, 'shutdowner.shutdown ' + str(x))
    dl = []
    try:
        shutdown_services()
        shutdown_interfaces()
        shutdown_local()
        shutdown_automats()
        shutdown_engine()
        shutdown_settings()
    except:
        lg.exc()
    # TODO: rework all shutdown() methods to return deferred objects
    return DeferredList(dl)


#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    global _Shutdowner
    if event is None:
        return _Shutdowner
    if _Shutdowner is None:
        _Shutdowner = Shutdowner(
            name='shutdowner',
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
        )
    if event is not None:
        _Shutdowner.event(event, *args, **kwargs)
    return _Shutdowner


#------------------------------------------------------------------------------


class Shutdowner(automat.Automat):
    """
    This is a state machine to manage a process of correctly finishing the
    BitDust software.
    """

    fast = True

    def init(self):
        self.flagApp = False
        self.flagReactor = False
        self.shutdown_param = None
        self.enableMemoryProfile = settings.enableMemoryProfile()

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        initializer.A('shutdowner.state', newstate)

    def A(self, event, *args, **kwargs):
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'INIT'
                self.flagApp = False
                self.flagReactor = False
        #---INIT---
        elif self.state == 'INIT':
            if event == 'stop':
                self.doSaveParam(*args, **kwargs)
                self.flagApp = True
            elif event == 'reactor-stopped':
                self.flagReactor = True
            elif event == 'ready' and self.flagReactor:
                self.state = 'FINISHED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'ready' and not self.flagReactor and self.flagApp:
                self.state = 'STOPPING'
                self.doShutdown(*args, **kwargs)
            elif event == 'ready' and not self.flagReactor and not self.flagApp:
                self.state = 'READY'
        #---READY---
        elif self.state == 'READY':
            if event == 'stop':
                self.state = 'STOPPING'
                self.doShutdown(*args, **kwargs)
            elif event == 'reactor-stopped':
                self.state = 'FINISHED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'block':
                self.state = 'BLOCKED'
        #---BLOCKED---
        elif self.state == 'BLOCKED':
            if event == 'stop':
                self.doSaveParam(*args, **kwargs)
                self.flagApp = True
            elif event == 'reactor-stopped':
                self.flagReactor = True
            elif event == 'unblock' and not self.flagReactor and not self.flagApp:
                self.state = 'READY'
            elif event == 'unblock' and not self.flagReactor and self.flagApp:
                self.state = 'STOPPING'
                self.doShutdown(*args, **kwargs)
            elif event == 'unblock' and self.flagReactor:
                self.state = 'FINISHED'
                self.doDestroyMe(*args, **kwargs)
        #---FINISHED---
        elif self.state == 'FINISHED':
            pass
        #---STOPPING---
        elif self.state == 'STOPPING':
            if event == 'reactor-stopped':
                self.state = 'FINISHED'
                self.doDestroyMe(*args, **kwargs)
        return None

    def doSaveParam(self, *args, **kwargs):
        self.shutdown_param = args[0]
        if _Debug:
            lg.out(_DebugLevel, 'shutdowner.doSaveParam %s' % str(self.shutdown_param))

    def doShutdown(self, *args, **kwargs):
        if _Debug:
            lg.out(_DebugLevel, 'shutdowner.doShutdown %d machines currently' % len(automat.objects()))

        param = args[0]
        if self.shutdown_param is not None:
            param = self.shutdown_param
        if not args or args[0] is None:
            param = 'exit'
        elif isinstance(args[0], six.string_types):
            param = args[0]
        if param not in ['exit', 'restart', 'restartnshow']:
            param = 'exit'
        if param == 'exit':
            self._shutdown_exit()
        elif param == 'restart':
            self._shutdown_restart()
        elif param == 'restartnshow':
            self._shutdown_restart('show')

    def doDestroyMe(self, *args, **kwargs):
        """
        Action method.
        """
        global _Shutdowner
        del _Shutdowner
        _Shutdowner = None
        self.destroy()
        if _Debug:
            lg.out(_DebugLevel, 'shutdowner.doDestroyMe %d machines left in memory:\n        %s' % (len(automat.objects()), '\n        '.join(['%d: %r' % (k, automat.by_index(k)) for k in automat.objects().keys()])))

        if self.enableMemoryProfile:
            try:
                from guppy import hpy  # @UnresolvedImport
                hp = hpy()
                hp.setrelheap()
                if _Debug:
                    lg.out(_DebugLevel, 'hp.heap():\n' + str(hp.heap()))
                    lg.out(_DebugLevel, 'hp.heap().byrcs:\n' + str(hp.heap().byrcs))
                    lg.out(_DebugLevel, 'hp.heap().byvia:\n' + str(hp.heap().byvia))
            except:
                if _Debug:
                    lg.out(_DebugLevel, 'guppy package is not installed')

    def _shutdown_restart(self, param=''):
        """
        Calls ``shutdown()`` method and stop the main reactor, then restart the
        program.
        """
        if _Debug:
            lg.out(_DebugLevel, 'shutdowner.shutdown_restart param=%s' % param)

        def do_restart(param):
            from bitdust.lib import misc
            from bitdust.system import bpio
            detach = False
            if bpio.Windows():
                detach = True
            misc.DoRestart(param, detach=detach)

        def shutdown_finished(x, param):
            if _Debug:
                lg.out(_DebugLevel, 'shutdowner.shutdown_finished want to stop the reactor')
            reactor.addSystemEventTrigger('after', 'shutdown', do_restart, param)  # @UndefinedVariable
            reactor.stop()  # @UndefinedVariable

        d = shutdown('restart')
        d.addBoth(shutdown_finished, param)

    def _shutdown_exit(self, x=None):
        """
        Calls ``shutdown()`` method and stop the main reactor, this will finish
        the program.
        """
        if _Debug:
            lg.out(_DebugLevel, 'shutdowner.shutdown_exit')

        def shutdown_reactor_stop(x=None):
            if _Debug:
                lg.out(_DebugLevel, 'shutdowner.shutdown_reactor_stop want to stop the reactor')
            reactor.stop()  # @UndefinedVariable

        d = shutdown(x)
        d.addBoth(shutdown_reactor_stop)



"""
Modularize the mining algo check

Bismuth Heavy3

From bismuth node
"""

import mmap
import os
import struct
import sys
import threading
from hashlib import sha224

from hmac_drbg import DRBG
from quantizer import quantize_ten
from decimal import Decimal
import regnet

from fork import Fork

fork = Fork()

__version__ = '0.1.4'

# print('Mining_Heavy3 v{}'.format(__version__))

RND_LEN = 0

MMAP = None
F = None

is_regnet = False
heavy = True


def read_int_from_map(map, index):
    return struct.unpack('I', map[4*index:4*index + 4])[0]


def anneal3(mmap, n):
    """
    Converts 224 bits number into annealed version, hexstring

    :param n: a 224 = 7x32 bits
    :return:  56 char in hex encoding.
    """
    # print('anneal3', threading.current_thread(), mmap, RND_LEN, id(RND_LEN))
    h7 = n & 0xffffffff
    n = n >> 32
    index = ((h7 & ~0x7) % RND_LEN)*4
    # f1 = struct.unpack('I', mmap[index:index + 4])[0]
    value = h7 ^ struct.unpack('I', mmap[index:index + 4])[0]
    res = '{:08x}'.format(value)
    for i in range(6):
        index += 4
        h = n & 0xffffffff
        n = n >> 32
        value = h ^ struct.unpack('I', mmap[index:index + 4])[0]
        res = '{:08x}'.format(value) + res
    return res


def anneal3_regnet(mmap, n):
    """
    Converts 224 bits number into annealed version, hexstring
    This uses a fake anneal for easier regtests

    :param n: a 224 = 7x32 bits
    :return:  56 char in hex encoding.
    """
    return '{:056x}'.format(n)


def bin_convert(string):
    return ''.join(format(ord(x), '8b').replace(' ', '0') for x in string)


def diffme_heavy3(pool_address, nonce, db_block_hash):
    # minimum possible diff
    diff = 1
    diff_result = 0
    hash224 = sha224((pool_address + nonce + db_block_hash).encode('utf-8')).digest()
    hash224 = int.from_bytes(hash224, 'big')
    if heavy or (not is_regnet):
        annealed_sha = anneal3(MMAP, hash224)
    else:
        annealed_sha = anneal3_regnet(MMAP, hash224)
    bin_annealed_sha = bin_convert(annealed_sha)
    mining_condition = bin_convert(db_block_hash)
    while mining_condition[:diff] in bin_annealed_sha:
        diff_result = diff
        diff += 1
    return diff_result


def check_block(block_height_new, miner_address, nonce, db_block_hash, diff0, received_timestamp, q_received_timestamp, q_db_timestamp_last, peer_ip='N/A', app_log=None):
    """
    Checks that the given block matches the mining algo.

    :param block_height_new:
    :param miner_address:
    :param nonce:
    :param db_block_hash:
    :param diff0:
    :param received_timestamp:
    :param q_received_timestamp:
    :param q_db_timestamp_last:
    :param peer_ip:
    :param app_log:
    :return:
    """
    try:
        if is_regnet:
            diff0 = regnet.REGNET_DIFF - 8

        # TODO: Verify!!!
        diff0 = 10

        real_diff = diffme_heavy3(miner_address, nonce, db_block_hash)
        diff_drop_time = Decimal(180)
        mining_condition = bin_convert(db_block_hash)[0:int(diff0)]
        # simplified comparison, no backwards mining
        if real_diff >= int(diff0):
            if app_log:
                app_log.info('Difficulty requirement satisfied for block {} from {}. {} >= {}'.format(block_height_new, peer_ip, real_diff, int(diff0)))
            diff_save = diff0

        elif Decimal(received_timestamp) > q_db_timestamp_last + Decimal(diff_drop_time):
            # uses block timestamp, don't merge with diff() for security reasons
            factor = 1
            time_difference = q_received_timestamp - q_db_timestamp_last
            diff_dropped = quantize_ten(diff0) + quantize_ten(1) - quantize_ten(time_difference/diff_drop_time)
            # Emergency diff drop
            if Decimal(received_timestamp) > q_db_timestamp_last + Decimal(2*diff_drop_time):
                factor = 10
                diff_dropped = quantize_ten(diff0) - quantize_ten(1) - quantize_ten(factor*(time_difference - 2*diff_drop_time)/diff_drop_time)

            if diff_dropped < 10:
                diff_dropped = 10
            if real_diff >= int(diff_dropped):
                if app_log:
                    app_log.info('Readjusted difficulty requirement satisfied for block {} from {}, {} >= {} (factor {})'.format(block_height_new, peer_ip, real_diff, int(diff_dropped), factor))
                diff_save = diff0
                # lie about what diff was matched not to mess up the diff algo
            else:
                raise ValueError('Readjusted difficulty too low for block {} from {}, {} should be at least {}'.format(block_height_new, peer_ip, real_diff, diff_dropped))
        else:
            raise ValueError('Difficulty {} too low for block {} from {}, should be at least {}'.format(real_diff, block_height_new, peer_ip, diff0))
        return diff_save
    except Exception as e:
        # Left for edge cases debug
        # print('MH3 check block', e)
        # import traceback
        # traceback.print_exc()
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        # print(exc_type, fname, exc_tb.tb_lineno)
        raise


def create_heavy3a(file_name='heavy3a.bin'):
    if (not heavy) and is_regnet:
        # print('Regnet, no heavy file')
        return
    # print('Creating Junction Noise file, this usually takes a few minutes...')
    gen = DRBG(b'Bismuth is a chemical element with symbol Bi and atomic number 83. It is a pentavalent post-transition metal and one of the pnictogens with chemical properties resembling its lighter homologs arsenic and antimony.')
    # Size in Gb - No more than 4Gb from a single seed
    GB = 1
    # Do not change chunk size, it would change the file content.
    CHUNK_SIZE = 1024*4  # time 3m20.990s
    COUNT = GB*1024*1024*1024 // CHUNK_SIZE
    with open(file_name, 'wb') as f:
        for chunks in range(COUNT):
            f.write(gen.generate(CHUNK_SIZE))


def mining_open(file_name='heavy3a.bin'):
    """
    Opens the Junction MMapped file
    """
    if (not heavy) and is_regnet:
        # print('Regnet, no heavy file to open')
        return
    global F
    global MMAP
    global RND_LEN
    if MMAP:
        print(f'Junction memory-map file already loaded in {threading.current_thread().name}')
        return
    print(f'Loading Junction memory-map file from {file_name} in {threading.current_thread().name}')
    if os.path.isfile(file_name):
        size = os.path.getsize(file_name)
        if size != 1073741824:
            # print('Invalid size of heavy file {}.'.format(file_name))
            try:
                os.remove(file_name)
                # print('Deleted, Will be re-created')
            except Exception as e:
                # print('mining_open', e)
                # sys.exit()
                raise e
    if not os.path.isfile(file_name):
        create_heavy3a(file_name)
    try:
        F = open(file_name, 'rb+')
        # memory-map the file, size 0 means whole file
        MMAP = mmap.mmap(F.fileno(), 0)
        RND_LEN = os.path.getsize(file_name) // 4
        if read_int_from_map(MMAP, 0) != 3786993664:
            raise ValueError('Wrong file: {}'.format(file_name))
        if read_int_from_map(MMAP, 1024) != 1742706086:
            raise ValueError('Wrong file: {}'.format(file_name))
    except Exception as e:
        # print('error while loading Junction file: {}'.format(e))
        # sys.exit()
        raise e
    # print('mining_open', file_name, threading.current_thread())


def mining_close():
    """
    Close the MMAP access, HAS to be called at end of program.
    """
    if (not heavy) and is_regnet:
        # print('Regnet, no heavy file to close')
        return
    global F
    global MMAP
    try:
        assert MMAP
        MMAP.close()
    except:
        pass
    MMAP = None

    try:
        assert F
        F.close()
    except:
        pass
    F = None



import re
import six
from six import PY2, b
if PY2:
    from itertools import ifilter as filter  # @UnresolvedImport
from functools import wraps
from twisted.web.resource import Resource, NoResource


class _FakeResource(Resource):
    _result = ''
    isLeaf = True

    def __init__(self, result):
        Resource.__init__(self)
        self._result = result

    def render(self, request):
        return self._result


def maybeResource(f):
    @wraps(f)
    def inner(*args, **kwargs):
        result = f(*args, **kwargs)
        if not isinstance(result, Resource):
            result = _FakeResource(result)
        return result

    return inner


class APIResource(Resource):

    _registry = None

    def __new__(cls, *args, **kwds):
        instance = super().__new__(cls, *args, **kwds)
        instance._registry = []
        for name in dir(instance):
            attribute = getattr(instance, name)
            annotations = getattr(attribute, '__txrestapi__', [])
            for annotation in annotations:
                method, regex = annotation
                instance.register(method, regex, attribute)
        return instance

    def __init__(self, *args, **kwargs):
        Resource.__init__(self, *args, **kwargs)
        if PY2:
            self._registry = []

    def _get_callback(self, request):
        filterf = lambda t: t[0] in (request.method, b('ALL'))
        path_to_check = getattr(request, '_remaining_path', request.path)
        if not isinstance(path_to_check, six.binary_type):
            path_to_check = path_to_check.encode()
        for m, r, cb in filter(filterf, self._registry):
            result = r.search(path_to_check)
            if result:
                request._remaining_path = path_to_check[result.span()[1]:]
                return cb, result.groupdict()
        return None, None

    def register(self, method, regex, callback):
        if not isinstance(regex, six.text_type):
            regex = regex.decode()
        self._registry.append((method, re.compile(regex), callback))

    def unregister(self, method=None, regex=None, callback=None):
        if regex is not None:
            if not isinstance(regex, six.text_type):
                regex = regex.decode()
            regex = re.compile(regex)
        for m, r, cb in self._registry[:]:
            if not method or (method and m == method):
                if not regex or (regex and r == regex):
                    if not callback or (callback and cb == callback):
                        self._registry.remove((m, r, cb))

    def getChild(self, name, request):
        r = self.children.get(name, None)
        if r is None:
            # Go into the thing
            callback, args = self._get_callback(request)
            if callback is None:
                return NoResource()
            else:
                return maybeResource(callback)(request, **args)
        else:
            return r



#!/usr/bin/python
# io_throttle.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (io_throttle.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: io_throttle.

When reconstructing a backup we don't want to take over everything
and make BitDust unresponsive by requesting 1000's of files at once
and make it so no other packets can go out.

This just tries to limit how much we are sending out or receiving at any time
so that we still have control.

Before requesting another file or sending another one out
I check to see how much stuff I have waiting.

Keep track of every supplier, store packets send/request in many queues.

TODO:
We probably want to be able to send not only to suppliers but to any contacts.
In future we can use that to do "overlay" communications to hide users.
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from six.moves import range

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 12

#------------------------------------------------------------------------------

import os
import sys
import time

#------------------------------------------------------------------------------

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in io_throttle.py')

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import misc
from bitdust.lib import nameurl
from bitdust.lib import packetid

from bitdust.main import settings
from bitdust.main import config

from bitdust.userid import my_id
from bitdust.userid import global_id
from bitdust.userid import id_url

from bitdust.p2p import commands
from bitdust.p2p import online_status

from bitdust.crypt import signed

from bitdust.transport import callback

#------------------------------------------------------------------------------

_IOThrottle = None
_PacketReportCallbackFunc = None

#------------------------------------------------------------------------------


def throttle():
    """
    Access method to the IO interface object.
    """
    global _IOThrottle
    if _IOThrottle is None:
        _IOThrottle = IOThrottle()
    return _IOThrottle


def init():
    """
    Init ``throttle()`` object and link with transports.
    This is the mechanism for sending and requesting files pieces.
    """
    if _Debug:
        lg.out(_DebugLevel, 'io_throttle.init')
    throttle()
    callback.add_queue_item_status_callback(OutboxStatus)
    callback.add_finish_file_sending_callback(FileSendingFinished)


def shutdown():
    if _Debug:
        lg.out(_DebugLevel, 'io_throttle.shutdown')
    callback.remove_finish_file_sending_callback(FileSendingFinished)
    callback.remove_queue_item_status_callback(OutboxStatus)
    throttle().DeleteBackupRequests('')
    throttle().DeleteBackupSendings('')
    throttle().DeleteSuppliers(list(throttle().supplierQueues.keys()))


#------------------------------------------------------------------------------


def SetPacketReportCallbackFunc(func):
    """
    You can pass a callback to catch a moment when some packet is
    added/removed.
    """
    global _PacketReportCallbackFunc
    _PacketReportCallbackFunc = func


def PacketReport(sendORrequest, supplier_idurl, packetID, result):
    """
    Called from other methods here to notify about packets events.
    """
    if _Debug:
        lg.out(_DebugLevel, 'io_throttle.PacketReport %s:%s %s to/from %s' % (sendORrequest, result, packetID, supplier_idurl))
    global _PacketReportCallbackFunc
    if _PacketReportCallbackFunc is not None:
        _PacketReportCallbackFunc(sendORrequest, supplier_idurl, packetID, result)


#------------------------------------------------------------------------------


def QueueSendFile(fileName, packetID, remoteID, ownerID, callOnAck=None, callOnFail=None):
    """
    Most used method - add an outgoing file to send to given remote peer.
    """
    return throttle().QueueSendFile(fileName, packetID, remoteID, ownerID, callOnAck, callOnFail)


def QueueRequestFile(callOnReceived, creatorID, packetID, ownerID, remoteID):
    """
    Place a request to download a single data packet from given remote supplier
    Remote user will verify our identity and decide to send the Data or not.
    Two scenarios possible when executing a `callOnReceived` callback:

        callOnReceived(newpacket, result)  or  callOnReceived(packetID, result)
    """
    return throttle().QueueRequestFile(callOnReceived, creatorID, packetID, ownerID, remoteID)


def DeleteBackupSendings(backupName):
    """
    Checks all send queues and search for packets with ``backupName`` in the
    packetID field.

    For example, this is used to remove old transfers if rebuilding
    process is restarted.
    """
    return throttle().DeleteBackupSendings(backupName)


def DeleteBackupRequests(backupName):
    """
    Checks all request queues and search for packets with ``backupName`` in the
    packetID field.

    For example, this is used to remove old requests if the restore
    process is aborted.
    """
    return throttle().DeleteBackupRequests(backupName)


def DeleteSuppliers(suppliers_IDURLs):
    """
    Erase the whole queue with this peer and remove him from throttle()
    completely.
    """
    return throttle().DeleteSuppliers(suppliers_IDURLs)


def DeleteAllSuppliers():
    return throttle().DeleteSuppliers(list(throttle().supplierQueues.keys()))


def OutboxStatus(pkt_out, status, error):
    return throttle().OutboxStatus(pkt_out, status, error)


def FileSendingFinished(pkt_out, item, status, size, error_message):
    return throttle().FileSendingFinished(pkt_out, item, status, size, error_message)


def IsSendingQueueEmpty():
    return throttle().IsSendingQueueEmpty()


def HasPacketInSendQueue(supplierIDURL, packetID):
    return throttle().HasPacketInSendQueue(supplierIDURL, packetID)


def HasPacketInRequestQueue(supplierIDURL, packetID):
    return throttle().HasPacketInRequestQueue(supplierIDURL, packetID)


def HasBackupIDInSendQueue(supplierIDURL, backupID):
    return throttle().HasBackupIDInSendQueue(supplierIDURL, backupID)


def HasBackupIDInRequestQueue(supplierIDURL, backupID):
    return throttle().HasBackupIDInRequestQueue(supplierIDURL, backupID)


def IsBackupSending(backupID):
    return throttle().IsBackupSending(backupID)


def OkToSend(supplierIDURL):
    return throttle().OkToSend(supplierIDURL)


def OkToRequest(supplierIDURL):
    return throttle().OkToRequest(supplierIDURL)


def GetSendQueueLength(supplierIDURL):
    return throttle().GetSendQueueLength(supplierIDURL)


def GetRequestQueueLength(supplierIDURL):
    return throttle().GetRequestQueueLength(supplierIDURL)


#------------------------------------------------------------------------------


class SupplierQueue:
    def __init__(self, supplierIdentity, creatorID, customerIDURL=None):
        self.customerIDURL = customerIDURL
        if self.customerIDURL is None:
            self.customerIDURL = my_id.getIDURL()
        self.creatorID = creatorID
        self.remoteID = supplierIdentity
        self.remoteName = nameurl.GetName(self.remoteID)

        # all sends we'll hold on to, only several will be active,
        # but will hold onto the next ones to be sent
        # active files
        self.fileSendMaxLength = config.conf().getInt('services/data-motion/supplier-sending-queue-size', 8)
        # an array of packetId, preserving first in first out,
        # of which the first maxLength are the "active" sends
        self.fileSendQueue = []
        # dictionary of FileUp's using packetId as index,
        # hold onto stuff sent and acked for some period as a history?
        self.fileSendDict = {}

        # all requests we'll hold on to,
        # only several will be active, but will hold onto the next ones to be sent
        # active requests
        self.fileRequestMaxLength = config.conf().getInt('services/data-motion/supplier-request-queue-size', 8)
        # an array of PacketIDs, preserving first in first out
        self.fileRequestQueue = []
        # FileDown's, indexed by PacketIDs
        self.fileRequestDict = {}

        self.shutdown = False

        self.ackedCount = 0
        self.failedCount = 0

        self.uploadingTimeoutCount = 0
        self.downloadingTimeoutCount = 0

        self._runSend = False
        self.sendTask = None
        self.sendTaskDelay = 0.1
        self.requestTask = None
        self.requestTaskDelay = 0.1

    #------------------------------------------------------------------------------

    def SupplierSendFile(self, fileName, packetID, ownerID, callOnAck=None, callOnFail=None):
        if self.shutdown:
            if _Debug:
                lg.out(_DebugLevel, 'io_throttle.SupplierSendFile finishing to %s, shutdown is True' % self.remoteName)
            if callOnFail is not None:
                reactor.callLater(0, callOnFail, self.remoteID, packetID, 'shutdown')  # @UndefinedVariable
            return False
        if online_status.isOffline(self.remoteID):
            if _Debug:
                lg.out(_DebugLevel, 'io_throttle.SupplierSendFile %s is offline, so packet %s is failed' % (self.remoteName, packetID))
            if callOnFail is not None:
                reactor.callLater(0, callOnFail, self.remoteID, packetID, 'offline')  # @UndefinedVariable
            return False
        if packetID in self.fileSendQueue:
            lg.warn('packet %s already in the queue for %s' % (packetID, self.remoteName))
            if callOnFail is not None:
                reactor.callLater(0, callOnFail, self.remoteID, packetID, 'in queue')  # @UndefinedVariable
            return False
        from bitdust.stream import file_up
        f_up = file_up.FileUp(
            self,
            fileName,
            packetID,
            self.remoteID,
            ownerID,
            callOnAck,
            callOnFail,
        )
        f_up.event('init')
        if _Debug:
            lg.out(_DebugLevel, 'io_throttle.SupplierSendFile %s to %s, %d queued items' % (packetID, self.remoteName, len(self.fileSendQueue)))
        self.DoSend()
        return True

    def StopAllSindings(self):
        for packetID in list(self.fileSendDict.keys()):
            f_up = self.fileSendDict.get(packetID)
            if f_up:
                if _Debug:
                    lg.args(_DebugLevel, packetID=packetID, obj=f_up, event='stop')
                f_up.event('stop')

    def DeleteBackupSendings(self, backupName):
        if self.shutdown:
            # if we're closing down this queue, don't do anything, but just stop all uploads
            self.StopAllSindings()
            return
        if _Debug:
            lg.args(_DebugLevel, backupName=backupName)
        packetsToRemove = set()
        for packetID in self.fileSendQueue:
            if (backupName and packetID.count(backupName)) or not backupName:
                packetsToRemove.add(packetID)
                if _Debug:
                    lg.out(_DebugLevel, 'io_throttle.DeleteBackupRequests %s from downloading queue' % packetID)
        for packetID in packetsToRemove:
            if packetID in self.fileSendDict:
                f_up = self.fileSendDict[packetID]
                f_up.event('stop')
                if _Debug:
                    lg.out(_DebugLevel, 'io_throttle.DeleteBackupRequests stopped %s in %s uploading queue, %d more items' % (packetID, self.remoteID, len(self.fileSendQueue)))
        if len(self.fileSendQueue) > 0:
            reactor.callLater(0, self.DoSend)  # @UndefinedVariable

    def OnFileSendAckReceived(self, newpacket, info):
        if self.shutdown:
            if _Debug:
                lg.out(_DebugLevel, 'io_throttle.OnFileSendAckReceived finishing to %s, shutdown is True' % self.remoteName)
            return
        if not newpacket and not info:
            lg.warn('packet timed out during responding')
            return
        if _Debug:
            lg.out(_DebugLevel, 'io_throttle.OnFileSendAckReceived with %r' % newpacket)
        self.ackedCount += 1
        packetID = global_id.CanonicalID(newpacket.PacketID)
        if packetID not in self.fileSendQueue:
            lg.warn('packet %s not in sending queue for %s' % (newpacket.PacketID, self.remoteName))
            return
        if packetID not in list(self.fileSendDict.keys()):
            lg.warn('packet %s not in sending dict for %s' % (newpacket.PacketID, self.remoteName))
            return
        f_up = self.fileSendDict[packetID]
        if newpacket.Command == commands.Ack():
            f_up.event('ack-received', newpacket)
        elif newpacket.Command == commands.Fail():
            f_up.event('fail-received', newpacket)
        else:
            raise Exception('wrong command received in response: %r' % newpacket)
        from bitdust.customer import supplier_connector
        sc = supplier_connector.by_idurl(newpacket.OwnerID)
        if sc:
            if newpacket.Command == commands.Ack():
                sc.automat('ack', newpacket)
            elif newpacket.Command == commands.Fail():
                sc.automat('fail', newpacket)
            else:
                raise Exception('incorrect packet type received: %r' % newpacket)
        else:
            lg.warn('supplier connector for %r not found' % newpacket.OwnerID)
        if _Debug:
            lg.out(_DebugLevel, 'io_throttle.OnFileSendAckReceived %s from %s, queue=%d' % (str(newpacket), self.remoteName, len(self.fileSendQueue)))

    def RunSend(self):
        if self._runSend:
            return
        self._runSend = True
        if _Debug:
            lg.out(_DebugLevel*2, 'io_throttle.RunSend  fileSendQueue=%d' % len(self.fileSendQueue))
        packetsToBeFailed = {}
        packetsToRemove = set()
        packetsSent = 0
        # let's check only beginning of the queue and try to process it.
        # once we finish and remove some items from the queue we can take more items
        for i in range(0, min(self.fileSendMaxLength, len(self.fileSendQueue))):
            try:
                packetID = self.fileSendQueue[i]
            except:
                lg.warn('item at position %d not exist in send queue' % i)
                continue

            f_up = self.fileSendDict[packetID]

            if f_up.state != 'IN_QUEUE':
                # we are sending that file at the moment
                packetsSent += 1
                # and we got ack
                if f_up.ackTime is None:
                    # if we did not get an ack yet we do not want to wait to long
                    if time.time() - f_up.sendTime > f_up.sendTimeout:
                        # so this packet is failed because no response for too long
                        packetsToBeFailed[packetID] = 'timeout'
                        lg.warn('uploading %r failed because of timeout %d src' % (packetID, f_up.sendTimeout))
                # this packet already in progress - check next one
                continue

            # the data file to send no longer exists - it is failed situation
            if not os.path.exists(f_up.fileName):
                lg.warn('file %s not exist' % (f_up.fileName))
                packetsToBeFailed[packetID] = 'not exist'
                continue

            # item is in the queue, but not started yet
            f_up.event('start')

        # process failed packets
        for packetID, why in packetsToBeFailed.items():
            f_up = self.fileSendDict[packetID]
            if why == 'timeout':
                f_up.event('timeout')
            elif why == 'not exist':
                f_up.event('file-not-exist')
            else:
                raise Exception('unknown result %r for %r' % (why, packetID))
        # remember results
        result = max(len(packetsToRemove), packetsSent)
        # erase temp lists
        del packetsToBeFailed
        del packetsToRemove
        self._runSend = False
        return result

    def SendingTask(self):
        if self.shutdown:
            self.StopAllSindings()
            return
        sends = self.RunSend()
        self.sendTaskDelay = misc.LoopAttenuation(self.sendTaskDelay, sends > 0, settings.MinimumSendingDelay(), settings.MaximumSendingDelay())
        # attenuation
        self.sendTask = reactor.callLater(self.sendTaskDelay, self.SendingTask)  # @UndefinedVariable

    def DoSend(self):
        if self.sendTask is None:
            self.SendingTask()
            return
        if self._runSend:
            return
        if self.sendTaskDelay > 1.0:
            self.sendTask.cancel()
            self.sendTask = None
            reactor.callLater(0, self.SendingTask)  # @UndefinedVariable

    #------------------------------------------------------------------------------

    def SupplierRequestFile(self, callOnReceived, creatorID, packetID, ownerID):
        if self.shutdown:
            if _Debug:
                lg.out(_DebugLevel, 'io_throttle.SupplierRequestFile finishing to %s, shutdown is True' % self.remoteName)
            if callOnReceived:
                reactor.callLater(0, callOnReceived, packetID, 'shutdown')  # @UndefinedVariable
            self.StopAllRequests()
            return False
        if packetID in self.fileRequestQueue:
            lg.warn('packet %s already in the queue for %s' % (packetID, self.remoteName))
            if callOnReceived:
                reactor.callLater(0, callOnReceived, packetID, 'in queue')  # @UndefinedVariable
            return False
        from bitdust.stream import file_down
        f_down = file_down.FileDown(self, callOnReceived, creatorID, packetID, ownerID, self.remoteID)
        f_down.event('init')
        if _Debug:
            lg.out(_DebugLevel, 'io_throttle.SupplierRequestFile %s from %s, %d queued items' % (packetID, self.remoteName, len(self.fileRequestQueue)))
        self.DoRequest()
        return True

    def StopAllRequests(self):
        for packetID in list(self.fileRequestDict.keys()):
            f_down = self.fileRequestDict.get(packetID)
            if f_down:
                if _Debug:
                    lg.args(_DebugLevel, packetID=packetID, obj=f_down, event='stop')
                f_down.event('stop')

    def DeleteBackupRequests(self, backupName):
        if self.shutdown:
            # if we're closing down this queue, don't do anything, but just stop all requests
            lg.warn('supplier queue is shutting down')
            self.StopAllRequests()
            return
        if _Debug:
            lg.out(_DebugLevel, 'io_throttle.DeleteBackupRequests  will cancel all requests for %s' % backupName)
        packetsToRemove = set()
        for packetID in self.fileRequestQueue:
            if (backupName and packetID.count(backupName)) or not backupName:
                packetsToRemove.add(packetID)
                if _Debug:
                    lg.out(_DebugLevel, 'io_throttle.DeleteBackupRequests %s from downloading queue' % packetID)
        for packetID in packetsToRemove:
            f_down = self.fileRequestDict.get(packetID)
            if f_down:
                f_down.event('stop')
                if _Debug:
                    lg.out(_DebugLevel, 'io_throttle.DeleteBackupRequests stopped %r in %s downloading queue, %d more items' % (packetID, self.remoteID, len(self.fileRequestQueue)))
            else:
                lg.warn('can not find %r in request queue' % packetID)
        if len(self.fileRequestQueue) > 0:
            reactor.callLater(0, self.DoRequest)  # @UndefinedVariable

    def OnDataReceived(self, newpacket, result):
        # we requested some data from a supplier, and just received it
        if self.shutdown:
            lg.warn('skip, supplier queue is shutting down')
            self.StopAllRequests()
            return
        if _Debug:
            lg.args(_DebugLevel, newpacket=newpacket, result=result, queue=len(self.fileRequestQueue), remoteName=self.remoteName)
        packetID = global_id.CanonicalID(newpacket.PacketID)
        if (packetID not in self.fileRequestQueue) or (packetID not in self.fileRequestDict):
            latest_idurl = global_id.NormalizeGlobalID(packetID, as_field=True)['idurl'].latest
            another_packetID = global_id.SubstitutePacketID(packetID, idurl=latest_idurl)
            if (another_packetID in self.fileRequestQueue) and (another_packetID in self.fileRequestDict):
                packetID = another_packetID
                lg.warn('found incoming %r with outdated packet id, corrected: %r' % (newpacket, another_packetID))
        if (packetID not in self.fileRequestQueue) or (packetID not in self.fileRequestDict):
            lg.err('unexpected %r received which is not in the downloading queue' % newpacket)
        else:
            f_down = self.fileRequestDict[packetID]
            if newpacket.Command == commands.Data():
                wrapped_packet = signed.Unserialize(newpacket.Payload)
                if not wrapped_packet or not wrapped_packet.Valid():
                    lg.err('incoming Data() packet is not valid')
                    f_down.event('fail-received', newpacket)
                    return
                f_down.event('valid-data-received', wrapped_packet)
            elif newpacket.Command == commands.Fail():
                f_down.event('fail-received', newpacket)
            else:
                lg.err('incorrect response command: %r' % newpacket)

    def RunRequest(self):
        packetsToRemove = {}
        for i in range(0, min(self.fileRequestMaxLength, len(self.fileRequestQueue))):
            packetID = self.fileRequestQueue[i]
            # must never happen, but just in case
            if packetID not in self.fileRequestDict:
                packetsToRemove[packetID] = 'broken'
                lg.err('file %r not found in downloading queue for %r' % (packetID, self.remoteID))
                continue
            f_down = self.fileRequestDict[packetID]
            if f_down.state == 'IN_QUEUE':
                customer, pathID = packetid.SplitPacketID(packetID)
                if os.path.exists(os.path.join(settings.getLocalBackupsDir(), customer, pathID)):
                    # we have the data file, no need to request it
                    packetsToRemove[packetID] = 'exist'
                else:
                    f_down.event('start')
        # remember requests results
        result = len(packetsToRemove)
        # remove finished requests
        for packetID, why in packetsToRemove.items():
            if _Debug:
                lg.out(_DebugLevel, 'io_throttle.RunRequest %r to be removed from [%s] downloading queue because %r, %d more items' % (packetID, self.remoteID, why, len(self.fileRequestQueue)))
            if packetID in self.fileRequestQueue:
                f_down = self.fileRequestDict[packetID]
                if why == 'exist':
                    f_down.event('file-already-exists')
                else:
                    lg.warn('unexpected result "%r" for %r in downloading queue for %s' % (why, packetID, self.remoteID))
                    f_down.event('stop')
            else:
                lg.warn('packet %r not found in request queue for [%s]' % (packetID, self.remoteID))
        del packetsToRemove
        if result:
            self.DoRequest()
        return result

    def RequestTask(self):
        if self.shutdown:
            self.StopAllRequests()
            return
        requests = self.RunRequest()
        self.requestTaskDelay = misc.LoopAttenuation(self.requestTaskDelay, requests > 0, settings.MinimumReceivingDelay(), settings.MaximumReceivingDelay())
        # attenuation
        self.requestTask = reactor.callLater(self.requestTaskDelay, self.RequestTask)  # @UndefinedVariable

    def DoRequest(self):
        if self.requestTask is None:
            self.RequestTask()
        else:
            if self.requestTaskDelay > 1.0:
                self.requestTask.cancel()
                self.requestTask = None
                self.RequestTask()

    #------------------------------------------------------------------------------

    def OnFileSendingFinished(self, pkt_out, item, status, size, error_message):
        if self.shutdown:
            lg.warn('skip, supplier queue is shutting down')
            return
        if not pkt_out.outpacket:
            lg.warn('skip, outpacket is already None')
            return
        packetID = global_id.CanonicalID(pkt_out.outpacket.PacketID)
        if status == 'finished':
            if pkt_out.outpacket.Command == commands.Retrieve():
                if packetID in self.fileRequestQueue:
                    f_down = self.fileRequestDict[packetID]
                    if _Debug:
                        lg.args(_DebugLevel, obj=f_down, status=status, packetID=packetID, event='retrieve-sent')
                    f_down.event('retrieve-sent', pkt_out.outpacket)
            elif pkt_out.outpacket.Command == commands.Data():
                if packetID in self.fileSendQueue:
                    f_up = self.fileSendDict[packetID]
                    if _Debug:
                        lg.args(_DebugLevel, obj=f_up, status=status, packetID=packetID, event='data-sent')
                    f_up.event('data-sent', pkt_out.outpacket)
        else:
            if pkt_out.outpacket.Command == commands.Retrieve():
                if packetID in self.fileRequestQueue:
                    if _Debug:
                        lg.dbg(_DebugLevel, 'packet %r is %r during downloading from %s' % (packetID, status, self.remoteID))
                    f_down = self.fileRequestDict[packetID]
                    f_down.event('request-failed')
            elif pkt_out.outpacket.Command == commands.Data():
                if packetID in self.fileSendQueue:
                    if _Debug:
                        lg.dbg(_DebugLevel, 'packet %r is %r during uploading to %s' % (packetID, status, self.remoteID))
                    f_up = self.fileSendDict[packetID]
                    f_up.event('sending-failed')

    def OutboxStatus(self, pkt_out, status, error):
        if self.shutdown:
            lg.warn('supplier queue is shutting down')
            return False
        packetID = global_id.CanonicalID(pkt_out.outpacket.PacketID)
        if status == 'finished':
            if pkt_out.outpacket.Command == commands.Data():
                if packetID in self.fileSendQueue:
                    f_up = self.fileSendDict[packetID]
                    if _Debug:
                        lg.args(_DebugLevel, obj=f_up, status=status, packetID=packetID, event='data-sent')
                    if error == 'unanswered':
                        f_up.event('timeout', pkt_out.outpacket)
                    else:
                        f_up.event('data-sent', pkt_out.outpacket)
                    return False
        else:
            if pkt_out.outpacket.Command == commands.Data():
                if packetID in self.fileSendQueue:
                    lg.warn('packet %r is %r during uploading to %s' % (packetID, status, self.remoteID))
                    f_up = self.fileSendDict[packetID]
                    f_up.event('sending-failed')
                    return False
        return False

    #------------------------------------------------------------------------------

    def RemoveSupplierWork(self):
        if _Debug:
            lg.out(_DebugLevel, 'io_throttle.RemoveSupplierWork for %r' % self.remoteID)
        self.DeleteBackupSendings(backupName=None)
        self.DeleteBackupRequests(backupName=None)

    #------------------------------------------------------------------------------

    def ListSendItems(self):
        return self.fileSendQueue

    def GetSendItem(self, packetID):
        return self.fileSendDict.get(packetID)

    def ListRequestItems(self):
        return self.fileRequestQueue

    def GetRequestItem(self, packetID):
        return self.fileRequestDict.get(packetID)

    def HasSendingFiles(self):
        return len(self.fileSendQueue) > 0

    def HasRequestedFiles(self):
        return len(self.fileRequestQueue) > 0

    def OkToSend(self):
        return len(self.fileSendQueue) < self.fileSendMaxLength

    def OkToRequest(self):
        return len(self.fileRequestQueue) < self.fileRequestMaxLength

    def GetSendQueueLength(self):
        return len(self.fileSendQueue)

    def GetRequestQueueLength(self):
        return len(self.fileRequestQueue)


#------------------------------------------------------------------------------


class IOThrottle:
    """
    All of the backup rebuilds will run their data requests through this
    So it gets throttled, also to reduce duplicate requests.
    """
    def __init__(self):
        self.creatorID = my_id.getIDURL()
        self.supplierQueues = {}
        self.paintFunc = None

    def GetSupplierQueue(self, supplierIDURL):
        supplierIDURL = id_url.field(supplierIDURL)
        return self.supplierQueues.get(supplierIDURL)

    def ListSupplierQueues(self):
        return list(self.supplierQueues.keys())

    def SetSupplierQueueCallbackFunc(self, func):
        self.paintFunc = func

    def DeleteSuppliers(self, suppliers_IDURLs):
        for supplierIDURL in id_url.to_list(suppliers_IDURLs):
            if supplierIDURL:
                if supplierIDURL in self.supplierQueues:
                    self.supplierQueues[supplierIDURL].RemoveSupplierWork()
                    del self.supplierQueues[supplierIDURL]

    def DeleteBackupSendings(self, backupName):
        # lg.out(10, 'io_throttle.DeleteBackupSendings for %s' % backupName)
        for supplierQueue in self.supplierQueues.values():
            supplierQueue.DeleteBackupSendings(backupName)

    def DeleteBackupRequests(self, backupName):
        # lg.out(10, 'io_throttle.DeleteBackupRequests for %s' % backupName)
        for supplierQueue in self.supplierQueues.values():
            supplierQueue.DeleteBackupRequests(backupName)

    def QueueSendFile(self, fileName, packetID, remoteID, ownerID, callOnAck=None, callOnFail=None):
        #out(10, "io_throttle.QueueSendFile %s to %s" % (packetID, nameurl.GetName(remoteID)))
        remoteID = id_url.field(remoteID)
        ownerID = id_url.field(ownerID)
        if not os.path.exists(fileName):
            lg.err('%s not exist' % fileName)
            if callOnFail is not None:
                reactor.callLater(.01, callOnFail, remoteID, packetID, 'not exist')  # @UndefinedVariable
            return False
        if remoteID not in list(self.supplierQueues.keys()):
            self.supplierQueues[remoteID] = SupplierQueue(remoteID, self.creatorID)
            lg.info('made a new sending queue for %s' % nameurl.GetName(remoteID))
        return self.supplierQueues[remoteID].SupplierSendFile(
            fileName,
            packetID,
            ownerID,
            callOnAck,
            callOnFail,
        )

    # return result in the callback: callOnReceived(packet or packetID, state)
    # state is: received, exist, in queue, shutdown
    def QueueRequestFile(self, callOnReceived, creatorID, packetID, ownerID, remoteID):
        # make sure that we don't actually already have the file
        remoteID = id_url.field(remoteID)
        ownerID = id_url.field(ownerID)
        creatorID = id_url.field(creatorID)
        if packetID != settings.BackupIndexFileName():
            customer, pathID = packetid.SplitPacketID(packetID)
            filename = os.path.join(settings.getLocalBackupsDir(), customer, pathID)
            if os.path.exists(filename):
                lg.warn('%s already exist ' % filename)
                if callOnReceived:
                    reactor.callLater(0, callOnReceived, packetID, 'exist')  # @UndefinedVariable
                return False
        if remoteID not in list(self.supplierQueues.keys()):
            # made a new queue for this man
            self.supplierQueues[remoteID] = SupplierQueue(remoteID, self.creatorID)
            lg.info('made a new receiving queue for %s' % nameurl.GetName(remoteID))
        # lg.out(10, "io_throttle.QueueRequestFile asking for %s from %s" % (packetID, nameurl.GetName(remoteID)))
        return self.supplierQueues[remoteID].SupplierRequestFile(callOnReceived, creatorID, packetID, ownerID)

    def OutboxStatus(self, pkt_out, status, error):
        """
        Called from outside to notify about single file sending result.
        """
        for supplierQueue in self.supplierQueues.values():
            if supplierQueue.OutboxStatus(pkt_out, status, error):
                return True
        return False

    def FileSendingFinished(self, pkt_out, item, status, size, error_message):
        for supplierQueue in self.supplierQueues.values():
            supplierQueue.OnFileSendingFinished(pkt_out, item, status, size, error_message)
        return False

    def IsSendingQueueEmpty(self):
        """
        Return True if all outgoing queues is empty, no sending at the moment.
        """
        for idurl in self.supplierQueues.keys():
            if self.supplierQueues[idurl].HasSendingFiles():
                # if _Debug:
                #     lg.out(_DebugLevel, 'io_throttle.IsSendingQueueEmpty   supplier %r has sending files:\n%r' % (
                #         idurl, self.supplierQueues[idurl].fileSendQueue))
                return False
        return True

    def IsRequestQueueEmpty(self):
        """
        Return True if all incoming queues is empty, no requests at the moment.
        """
        for idurl in self.supplierQueues.keys():
            if not self.supplierQueues[idurl].HasRequestedFiles():
                if _Debug:
                    lg.out(_DebugLevel, 'io_throttle.IsRequestQueueEmpty   supplier %r has requested files:\n%r' % (idurl, self.supplierQueues[idurl].fileRequestQueue))
                return False
        return True

    def HasPacketInSendQueue(self, supplierIDURL, packetID):
        """
        Return True if that packet is found in the sending queue to given
        remote peer.
        """
        supplierIDURL = id_url.field(supplierIDURL)
        if supplierIDURL not in self.supplierQueues:
            return False
        return packetID in self.supplierQueues[supplierIDURL].fileSendDict

    def HasPacketInRequestQueue(self, supplierIDURL, packetID):
        """
        Return True if that packet is found in the request queue from given
        remote peer.
        """
        supplierIDURL = id_url.field(supplierIDURL)
        if supplierIDURL not in self.supplierQueues:
            return False
        return packetID in self.supplierQueues[supplierIDURL].fileRequestDict

    def HasBackupIDInSendQueue(self, supplierIDURL, backupID):
        """
        Same to ``HasPacketInSendQueue()``, but looks for packets for the whole
        backup, not just a single packet .
        """
        supplierIDURL = id_url.field(supplierIDURL)
        if supplierIDURL not in self.supplierQueues:
            return False
        for packetID in self.supplierQueues[supplierIDURL].fileSendDict.keys():
            if packetID.count(backupID):
                return True
        return False

    def HasBackupIDInRequestQueue(self, supplierIDURL, backupID):
        """
        Same to ``HasPacketInRequestQueue()``, but looks for packets for the
        whole backup, not just a single packet .
        """
        supplierIDURL = id_url.field(supplierIDURL)
        if supplierIDURL not in self.supplierQueues:
            return False
        for packetID in self.supplierQueues[supplierIDURL].fileRequestDict.keys():
            if packetID.count(backupID):
                return True
        return False

    def IsBackupSending(self, backupID):
        """
        Return True if some packets for given backup is found in the sending
        queues.
        """
        for supplierIDURL in self.supplierQueues.keys():
            if self.HasBackupIDInSendQueue(supplierIDURL, backupID):
                return True
        return False

    def IsBackupRequesting(self, backupID):
        """
        Return True if some packets for given backup is found in the request
        queues.
        """
        for supplierIDURL in self.supplierQueues.keys():
            if self.HasBackupIDInRequestQueue(supplierIDURL, backupID):
                return True
        return False

    def OkToSend(self, supplierIDURL):
        """
        The maximum size of any queue is limited, if this limit is not reached
        yet you can put more files to send to that remote user.

        This method return True if you can put more outgoing files to
        that man in the ``throttle()``.
        """
        supplierIDURL = id_url.field(supplierIDURL)
        if supplierIDURL not in self.supplierQueues:
            # no queue opened to this man, so the queue is ready
            return True
        return self.supplierQueues[supplierIDURL].OkToSend()

    def GetRequestQueueLength(self, supplierIDURL):
        """
        Return number of requested packets from given user.
        """
        supplierIDURL = id_url.field(supplierIDURL)
        if supplierIDURL not in self.supplierQueues:
            # no queue opened to this man, so length is zero
            return 0
        return self.supplierQueues[supplierIDURL].GetRequestQueueLength()

    def GetSendQueueLength(self, supplierIDURL):
        """
        Return number of packets sent to this guy.
        """
        supplierIDURL = id_url.field(supplierIDURL)
        if supplierIDURL not in self.supplierQueues:
            # no queue opened to this man, so length is zero
            return 0
        return self.supplierQueues[supplierIDURL].GetSendQueueLength()



#!/usr/bin/python
# identitydb.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (identitydb.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
"""
.. module:: identitydb.

Here is a simple1 database for identities cache. Also keep track of
changing identities sources and maintain a several "index" dictionaries
to speed up processes.
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
import os
import time

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio

from bitdust.main import settings

from bitdust.lib import nameurl

from bitdust.userid import identity
from bitdust.userid import id_url

#------------------------------------------------------------------------------

# Dictionary cache of identities - lookup by primary url
# global dictionary of identities in this file
# indexed with urls and contains identity objects
_IdentityCache = {}
_IdentityCacheIDs = {}
_IdentityCacheCounter = 0
_IdentityCacheModifiedTime = {}
_Contact2IDURL = {}
_IDURL2Contacts = {}
_IPPort2IDURL = {}
_LocalIPs = {}
_IdentityCacheUpdatedCallbacks = []

#------------------------------------------------------------------------------


def cache():
    global _IdentityCache
    return _IdentityCache


def cache_ids():
    global _IdentityCacheIDs
    return _IdentityCacheIDs


def cache_contacts():
    global _Contact2IDURL
    return _Contact2IDURL


#------------------------------------------------------------------------------


def init():
    """
    Need to call before all other methods.

    Check to exist and create a folder to keep all cached identities.
    """
    if _Debug:
        lg.out(_DebugLevel, 'identitydb.init')
    id_cache_dir = settings.IdentityCacheDir()
    if not os.path.exists(id_cache_dir):
        if _Debug:
            lg.out(_DebugLevel, 'identitydb.init create folder %r' % id_cache_dir)
        bpio._dir_make(id_cache_dir)
    # make sure to read and cache all known identities at startup
    for id_filename in os.listdir(id_cache_dir):
        idurl = nameurl.FilenameUrl(id_filename)
        get_ident(idurl)


def shutdown():
    if _Debug:
        lg.out(_DebugLevel, 'identitydb.shutdown')


#------------------------------------------------------------------------------


def clear(exclude_list=None):
    """
    Clear the database, indexes and cached files from disk.
    """
    global _Contact2IDURL
    global _IPPort2IDURL
    global _IDURL2Contacts
    global _IdentityCache
    global _IdentityCacheIDs
    global _IdentityCacheModifiedTime
    if _Debug:
        lg.out(_DebugLevel, 'identitydb.clear')
    _IdentityCache.clear()
    _IdentityCacheIDs.clear()
    _IdentityCacheModifiedTime.clear()
    _Contact2IDURL.clear()
    _IPPort2IDURL.clear()
    _IDURL2Contacts.clear()
    iddir = settings.IdentityCacheDir()
    if not os.path.exists(iddir):
        return
    for file_name in os.listdir(iddir):
        path = os.path.join(iddir, file_name)
        if not os.access(path, os.W_OK):
            continue
        if exclude_list:
            idurl = nameurl.FilenameUrl(file_name)
            if idurl in exclude_list:
                continue
        os.remove(path)
        if _Debug:
            lg.out(_DebugLevel, 'identitydb.clear remove ' + path)
    fire_cache_updated_callbacks()


def size():
    """
    Return a number of items in the database.
    """
    global _IdentityCache
    return len(_IdentityCache)


def has_idurl(idurl):
    """
    Return True if that IDURL already cached.
    """
    global _IdentityCache
    return id_url.to_original(idurl) in _IdentityCache


def has_file(idurl):
    idurl = id_url.to_original(idurl)
    try:
        partfilename = nameurl.UrlFilename(idurl)
    except:
        lg.err('idurl %r is not correct' % idurl)
        return None
    filename = os.path.join(settings.IdentityCacheDir(), partfilename)
    return os.path.exists(filename)


def idset(idurl, id_obj):
    """
    Important method - need to call that to update indexes.
    """
    global _Contact2IDURL
    global _IDURL2Contacts
    global _IPPort2IDURL
    global _IdentityCache
    global _IdentityCacheIDs
    global _IdentityCacheCounter
    global _IdentityCacheModifiedTime
    idurl = id_url.to_original(idurl)
    if not has_idurl(idurl):
        if _Debug:
            lg.out(_DebugLevel, 'identitydb.idset new identity: %r' % idurl)
    _IdentityCache[idurl] = id_obj
    _IdentityCacheModifiedTime[idurl] = time.time()
    identid = _IdentityCacheIDs.get(idurl, None)
    if identid is None:
        identid = _IdentityCacheCounter
        _IdentityCacheCounter += 1
        _IdentityCacheIDs[idurl] = identid
    for contact in id_obj.getContacts():
        if contact not in _Contact2IDURL:
            _Contact2IDURL[contact] = set()
        # else:
        #     if len(_Contact2IDURL[contact]) >= 1 and idurl not in _Contact2IDURL[contact]:
        #         lg.warn('another user have same contact: ' + str(list(_Contact2IDURL[contact])))
        _Contact2IDURL[contact].add(idurl)
        if idurl not in _IDURL2Contacts:
            _IDURL2Contacts[idurl] = set()
        _IDURL2Contacts[idurl].add(contact)
        try:
            proto, host, port, fname = nameurl.UrlParse(contact)
            ipport = (host, int(port))
            _IPPort2IDURL[ipport] = idurl
        except:
            pass
    # TODO: when identity contacts changed - need to remove old items from _Contact2IDURL
    fire_cache_updated_callbacks(single_item=(identid, idurl, id_obj))
    if _Debug:
        lg.out(_DebugLevel, 'identitydb.idset %r' % idurl)
    # now make sure we properly handle changes in the sources of that identity
    try:
        id_url.identity_cached(id_obj)
    except:
        lg.exc()


def idget(idurl):
    """
    Get identity from cache.
    """
    global _IdentityCache
    idurl = id_url.to_original(idurl)
    return _IdentityCache.get(idurl, None)


def idremove(idurl):
    """
    Remove identity from cache, also update indexes.

    Not remove local file.
    """
    global _IdentityCache
    global _IdentityCacheIDs
    global _IdentityCacheModifiedTime
    global _Contact2IDURL
    global _IDURL2Contacts
    global _IPPort2IDURL
    idurl = id_url.to_original(idurl)
    idobj = _IdentityCache.pop(idurl, None)
    identid = _IdentityCacheIDs.pop(idurl, None)
    _IdentityCacheModifiedTime.pop(idurl, None)
    _IDURL2Contacts.pop(idurl, None)
    if idobj is not None:
        for contact in idobj.getContacts():
            _Contact2IDURL.pop(contact, None)
            try:
                proto, host, port, fname = nameurl.UrlParse(contact)
                ipport = (host, int(port))
                _IPPort2IDURL.pop(ipport, None)
            except:
                pass
    fire_cache_updated_callbacks(single_item=(identid, None, None))
    if _Debug:
        lg.args(_DebugLevel, idurl=idurl)
    return idobj


def idcontacts(idurl):
    """
    A fast way to get identity contacts.
    """
    global _IDURL2Contacts
    idurl = id_url.to_original(idurl)
    return list(_IDURL2Contacts.get(idurl, set()))


def get_ident(idurl):
    """
    A smart way to get identity from cache.

    If not cached in memory but found locally - read it from disk.
    """
    idurl = id_url.to_original(idurl)
    if has_idurl(idurl):
        return idget(idurl)
    try:
        partfilename = nameurl.UrlFilename(idurl)
    except:
        if _Debug:
            lg.out(_DebugLevel, 'identitydb.get_ident ERROR %r is incorrect' % idurl)
        return None
    if not partfilename:
        if _Debug:
            lg.out(_DebugLevel, 'identitydb.get_ident ERROR %r is empty' % idurl)
        return None
    filename = os.path.join(settings.IdentityCacheDir(), partfilename)
    if not os.path.exists(filename):
        if _Debug:
            lg.out(_DebugLevel, 'identitydb.get_ident file %r not exist' % os.path.basename(filename))
        return None
    idxml = bpio.ReadTextFile(filename)
    if not idxml:
        if _Debug:
            lg.out(_DebugLevel, 'identitydb.get_ident %s not found' % nameurl.GetName(idurl))
        return None
    try:
        idobj = identity.identity(xmlsrc=idxml)
        idurl_orig = idobj.getIDURL()
    except:
        lg.exc()
        return None
    if idurl == idurl_orig.original():
        idset(idurl, idobj)
        return idobj
    lg.err('not found identity object idurl=%r idurl_orig=%r' % (idurl, idurl_orig))
    return None


def get_filename(idurl):
    idurl = id_url.to_original(idurl)
    try:
        partfilename = nameurl.UrlFilename(idurl)
    except:
        lg.err('idurl %r is not correct' % idurl)
        return None
    return os.path.join(settings.IdentityCacheDir(), partfilename)


def get_idurls_by_contact(contact):
    """
    Use index dictionary to get IDURL with given contact.
    """
    global _Contact2IDURL
    return list(_Contact2IDURL.get(contact, set()))


def get_idurl_by_ip_port(ip, port):
    """
    Use index dictionary to get IDURL by IP and PORT.
    """
    global _IPPort2IDURL
    return _IPPort2IDURL.get((ip, int(port)), None)


def update(idurl, xml_src):
    """
    This is a correct method to update an identity in the local cache.

    PREPRO need to check that date or version is after old one so not
    vulnerable to replay attacks.
    """
    idurl = id_url.to_original(idurl)
    try:
        newid = identity.identity(xmlsrc=xml_src)
    except:
        lg.exc()
        return False

    if not newid.isCorrect():
        lg.warn('incorrect identity : %r' % idurl)
        return False

    try:
        if not newid.Valid():
            lg.warn('identity not valid : %r' % idurl)
            return False
    except:
        lg.exc()
        return False

    new_idurl = newid.getIDURL(as_original=True)
    if idurl != new_idurl:
        lg.warn('original IDURL is not matching, updated: %r -> %r' % (idurl, new_idurl))
        idurl = new_idurl

    filename = os.path.join(settings.IdentityCacheDir(), nameurl.UrlFilename(idurl))
    if os.path.exists(filename):
        oldidentityxml = bpio.ReadTextFile(filename)
        oldidentity = identity.identity(xmlsrc=oldidentityxml)

        if oldidentity.publickey != newid.publickey:
            if _Debug:
                lg.args(_DebugLevel, old=oldidentity.publickey, new=newid.publickey)
            lg.exc(exc_value=Exception('received public key from %s does not match with already cached identity' % idurl))
            return False

        if oldidentity.signature != newid.signature:
            if _Debug:
                lg.out(_DebugLevel, 'identitydb.update have new data for %r' % nameurl.GetName(idurl))
        else:
            idset(idurl, newid)
            return True
    # public keys are matching so we can update it
    bpio.WriteTextFile(filename, xml_src)
    idset(idurl, newid)
    return True


def remove(idurl):
    """
    Top method to remove identity from cache - also remove local file.
    """
    idurl = id_url.to_original(idurl)
    filename = os.path.join(settings.IdentityCacheDir(), nameurl.UrlFilename(idurl))
    if os.path.isfile(filename):
        if _Debug:
            lg.out(_DebugLevel, 'identitydb.remove file %r' % filename)
        try:
            os.remove(filename)
        except:
            lg.exc()
    idremove(idurl)
    return True


def update_local_ips_dict(local_ips_dict):
    """
    This method intended to maintain a local IP's index.
    """
    global _LocalIPs
    # _LocalIPs.clear()
    # _LocalIPs = local_ips_dict
    _LocalIPs.update(local_ips_dict)


def get_local_ip(idurl):
    """
    This is to get a local IP of some user from the index.
    """
    global _LocalIPs
    idurl = id_url.to_original(idurl)
    return _LocalIPs.get(idurl, None)


def has_local_ip(idurl):
    """
    To check for some known local IP of given user.
    """
    global _LocalIPs
    idurl = id_url.to_original(idurl)
    return idurl in _LocalIPs


def search_local_ip(ip):
    """
    Search all index for given local IP and return a first found idurl.
    """
    global _LocalIPs
    for idurl, localip in _LocalIPs.items():
        if localip == ip:
            return idurl
    return None


def get_last_modified_time(idurl):
    global _IdentityCacheModifiedTime
    idurl = id_url.to_original(idurl)
    return _IdentityCacheModifiedTime.get(idurl, None)


#------------------------------------------------------------------------------


def print_id(idurl):
    """
    For debug purposes.
    """
    idurl = id_url.to_original(idurl)
    if has_idurl(idurl):
        idForKey = get_ident(idurl)
        if _Debug:
            lg.out(_DebugLevel, str(idForKey.getSources(as_originals=True)))
            lg.out(_DebugLevel, str(idForKey.contacts))
            lg.out(_DebugLevel, str(idForKey.publickey))
            lg.out(_DebugLevel, str(idForKey.signature))


def print_keys():
    """
    For debug purposes.
    """
    global _IdentityCache
    for key in _IdentityCache.keys():
        if _Debug:
            lg.out(_DebugLevel, '%d: %r' % (_IdentityCacheIDs[key], key))


def print_cache():
    """
    For debug purposes.
    """
    global _IdentityCache
    for key in _IdentityCache.keys():
        if _Debug:
            lg.out(_DebugLevel, '---------------------')
        print_id(key)


#------------------------------------------------------------------------------


def AddCacheUpdatedCallback(cb):
    global _IdentityCacheUpdatedCallbacks
    _IdentityCacheUpdatedCallbacks.append(cb)


def RemoveCacheUpdatedCallback(cb):
    global _IdentityCacheUpdatedCallbacks
    if cb in _IdentityCacheUpdatedCallbacks:
        _IdentityCacheUpdatedCallbacks.remove(cb)


def fire_cache_updated_callbacks(single_item=None):
    global _IdentityCacheUpdatedCallbacks
    for cb in _IdentityCacheUpdatedCallbacks:
        cb(cache_ids(), cache(), single_item)



#!/usr/bin/python
# service_private_groups.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_private_groups.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_private_groups
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return PrivateGroupsService()


class PrivateGroupsService(LocalService):

    service_name = 'service_private_groups'
    config_path = 'services/private-groups/enabled'

    def dependent_on(self):
        return [
            'service_my_data',
            'service_private_messages',
        ]

    def start(self):
        from bitdust.main import events
        from bitdust.access import groups
        from bitdust.access import group_participant
        groups.init()
        events.add_subscriber(self._on_supplier_modified, 'supplier-modified')
        group_participant.start_group_participants()
        events.add_subscriber(groups.on_identity_url_changed, 'identity-url-changed')
        return True

    def stop(self):
        from bitdust.main import events
        from bitdust.access import groups
        from bitdust.access import group_participant
        events.remove_subscriber(groups.on_identity_url_changed, 'identity-url-changed')
        group_participant.shutdown_group_participants()
        events.remove_subscriber(self._on_supplier_modified, 'supplier-modified')
        groups.shutdown()
        return True

    def health_check(self):
        # TODO: probably at least one queue must be connected if service is enabled
        return True

    def _on_supplier_modified(self, evt):
        from bitdust.logs import lg
        from bitdust.access import key_ring
        from bitdust.crypt import my_keys
        from bitdust.userid import global_id
        from bitdust.userid import my_id
        if evt.data['new_idurl']:
            my_keys_to_be_republished = []
            for key_id in my_keys.known_keys():
                if not key_id.startswith('group_'):
                    continue
                _glob_id = global_id.NormalizeGlobalID(key_id)
                if _glob_id['idurl'] == my_id.getIDURL():
                    # only send public keys of my own groups
                    my_keys_to_be_republished.append(key_id)
            for group_key_id in my_keys_to_be_republished:
                d = key_ring.transfer_key(group_key_id, trusted_idurl=evt.data['new_idurl'], include_private=False, include_signature=False)
                d.addErrback(lambda *a: lg.err('transfer key failed: %s' % str(*a)))



#!/usr/bin/env python
# id_server.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (id_server.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: id_server.

.. role:: red
BitDust id_server() Automat

EVENTS:
    * :red:`incoming-identity-file`
    * :red:`init`
    * :red:`server-down`
    * :red:`shutdown`
    * :red:`start`
    * :red:`stop`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 14

#------------------------------------------------------------------------------

from io import BytesIO

#------------------------------------------------------------------------------

import os
import sys
import struct

from twisted.internet import reactor  # @UnresolvedImport
from twisted.internet.protocol import ServerFactory
from twisted.internet.defer import Deferred, DeferredList
from twisted.protocols import basic
from twisted.web import server, resource, static

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..', '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio
from bitdust.system import tmpfile

from bitdust.automats import automat

from bitdust.lib import strng
from bitdust.lib import nameurl
from bitdust.lib import misc
from bitdust.lib import net_misc

from bitdust.interface import web_html_template

from bitdust.main import settings

from bitdust.userid import identity
from bitdust.userid import known_servers

#------------------------------------------------------------------------------

_IdServer = None

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _IdServer
    if _IdServer is None:
        # set automat name and starting state here
        _IdServer = IdServer(
            name='id_server',
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
        )
    if event is None:
        return _IdServer
    _IdServer.automat(event, *args, **kwargs)


class IdServer(automat.Automat):

    """
    This class implements all the functionality of the ``id_server()`` state
    machine.
    """

    def init(self):
        """
        Method to initialize additional variables and flags at creation of the
        state machine.
        """
        self.web_listener = None
        self.tcp_listener = None
        self.web_port = settings.IdentityWebPort()
        self.tcp_port = settings.IdentityServerPort()
        self.hostname = ''

    def A(self, event, *args, **kwargs):
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'STOPPED'
                self.doInit(*args, **kwargs)
        #---LISTEN---
        elif self.state == 'LISTEN':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doSetDown(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'incoming-identity-file':
                self.doCheckAndSaveIdentity(*args, **kwargs)
            elif event == 'stop':
                self.state = 'DOWN'
                self.Restart = False
                self.doSetDown(*args, **kwargs)
        #---STOPPED---
        elif self.state == 'STOPPED':
            if event == 'start':
                self.state = 'LISTEN'
                self.doSetUp(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        #---DOWN---
        elif self.state == 'DOWN':
            if event == 'server-down' and self.Restart:
                self.state = 'LISTEN'
                self.doSetUp(*args, **kwargs)
            elif event == 'start':
                self.Restart = True
            elif event == 'server-down' and not self.Restart:
                self.state = 'STOPPED'
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        return None

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        self.web_port, self.tcp_port = args[0]

    def doSetUp(self, *args, **kwargs):
        """
        Action method.
        """
        self.hostname = settings.getIdServerHost()
        if self.hostname == '':
            self.hostname = strng.to_bin(misc.readExternalIP())
        if self.hostname == '':
            self.hostname = net_misc.getLocalIp()
        if _Debug:
            lg.out(_DebugLevel, 'id_server.doSetUp hostname=%s' % strng.to_text(self.hostname))
        if not os.path.isdir(settings.IdentityServerDir()):
            os.makedirs(settings.IdentityServerDir())
            if _Debug:
                lg.out(_DebugLevel, '            created a folder %s' % settings.IdentityServerDir())
        root = WebRoot()
        root.putChild(b'', WebMainPage())
        try:
            self.tcp_listener = reactor.listenTCP(self.tcp_port, IdServerFactory())  # @UndefinedVariable
            if _Debug:
                lg.out(_DebugLevel, '            identity server listen on TCP port %d started' % (self.tcp_port))
        except:
            if _Debug:
                lg.out(_DebugLevel, 'id_server.set_up ERROR exception trying to listen on port ' + str(self.tcp_port))
            lg.exc()
        try:
            self.web_listener = reactor.listenTCP(self.web_port, server.Site(root))  # @UndefinedVariable
            if _Debug:
                lg.out(_DebugLevel, '            have started web server at port %d   hostname=%s' % (self.web_port, strng.to_text(self.hostname)))
        except:
            if _Debug:
                lg.out(_DebugLevel, 'id_server.set_up ERROR exception trying to listen on port ' + str(self.web_port))
            lg.exc()

    def doSetDown(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'id_server.doSetDown')
        shutlist = []
        if self.web_listener:
            d = self.web_listener.stopListening()
            if d:
                shutlist.append(d)
            if _Debug:
                lg.out(_DebugLevel, '            stopped web listener')
        if self.tcp_listener:
            d = self.tcp_listener.stopListening()
            if d:
                shutlist.append(d)
            if _Debug:
                lg.out(_DebugLevel, '            stopped TCP listener')
        self.web_listener = None
        self.tcp_listener = None
        DeferredList(shutlist).addBoth(lambda x: self.automat('server-down'))

    def doCheckAndSaveIdentity(self, *args, **kwargs):
        """
        Action method.
        """
        self._save_identity(args[0])

    def doDestroyMe(self, *args, **kwargs):
        """
        Action method.
        """
        self.destroy()
        global _IdServer
        _IdServer = None
        if args and args[0] and len(args[0]) > 0 and isinstance(args[0][-1], Deferred):
            args[0][-1].callback(True)

    def _save_identity(self, inputfilename):
        """
        """
        if _Debug:
            lg.out(_DebugLevel, 'id_server._save_identity ' + inputfilename)
        if os.path.getsize(inputfilename) > 50000:
            lg.warn('input file too big - ignoring')
            tmpfile.erase('idsrv', inputfilename, 'input file too big')
            return
        newxml = bpio.ReadTextFile(inputfilename)
        if not newxml or len(newxml.strip()) < 500:
            lg.warn('input file too small - ignoring')
            tmpfile.erase('idsrv', inputfilename, 'input file too small')
            return
        if not newxml.startswith('<?xml version="1.0" encoding="utf-8"?>'):
            lg.warn('input file is not an XML - ignoring')
            tmpfile.erase('idsrv', inputfilename, 'input file is not an XML')
            return
        try:
            newidentity = identity.identity(xmlsrc=newxml)
        except:
            lg.warn('input file is wrong - ignoring')
            tmpfile.erase('idsrv', inputfilename, 'input file is wrong')
            return
        tmpfile.erase('idsrv', inputfilename, 'id received')
        if not newidentity.isCorrect():
            lg.warn('has non-Correct identity')
            return
        if not newidentity.Valid():
            lg.warn('has non-Valid identity')
            return
        matchid = b''
        for idurl_bin in newidentity.getSources(as_originals=True):
            protocol, host, port, filename = nameurl.UrlParse(idurl_bin)
            if strng.to_text(host) == strng.to_text(self.hostname):
                if _Debug:
                    lg.out(_DebugLevel, 'id_server._save_identity found match for us')
                matchid = idurl_bin
                break
        if not matchid:
            lg.warn('identity is not for this nameserver sources: %r' % newidentity.getSources(as_originals=True))
            return
        protocol, host, port, filename = nameurl.UrlParse(matchid)
        name, justxml = filename.split('.')
        # SECURITY check that name is simple
        if justxml != 'xml':
            lg.warn('identity name ' + filename)
            return
        if len(name) > settings.MaximumUsernameLength():
            lg.warn('identity name ' + filename)
            return
        if len(name) < settings.MinimumUsernameLength():
            lg.warn('identity name ' + filename)
            return
        for c in name:
            if c not in settings.LegalUsernameChars():
                lg.warn('identity name ' + filename)
                return
        localfilename = os.path.join(settings.IdentityServerDir(), filename)
        oldxml = ''
        # need to make sure id was not already used by different key - which would mean someone is trying to steal identity
        if os.path.exists(localfilename):
            if _Debug:
                lg.out(_DebugLevel, 'id_server._save_identity was already an identity with this name ' + localfilename)
            oldxml = bpio.ReadTextFile(localfilename)
            oldidentity = identity.identity(xmlsrc=oldxml)
            if oldidentity.publickey != newidentity.publickey:
                lg.warn('new public key does not match old ' + localfilename)
                return
        if newxml != oldxml:
            if not os.path.exists(localfilename):
                if _Debug:
                    lg.out(_DebugLevel, 'id_server._save_identity will save NEW Identity: ' + filename)
            bpio.WriteTextFile(localfilename, newxml)


#------------------------------------------------------------------------------


class IdServerProtocol(basic.Int32StringReceiver):

    def __init__(self):
        self.fpath = None  # string with path/filename
        self.fin = None  # integer file descriptor like os.open() returns
        self.received = 0

    def disconnect(self):
        try:
            self.transport.stopListening()
        except:
            try:
                self.transport.loseConnection()
            except:
                lg.exc()

    def connectionMade(self):
        """
        """

    def stringReceived(self, data):
        try:
            version = strng.to_bin(data[0:1])
            command = strng.to_bin(data[1:2])
            payload = data[2:]
        except:
            self.disconnect()
            lg.exc()
            lg.warn('incorrect data from %s\n' % str(self.transport.getPeer()))
            return
        if command == b'a':
            lg.warn('ignored incoming "abort" packet from %s' % str(self.transport.getPeer()))
            return
        if command == b'h':
            self.sendString(version + b'wid-server:' + strng.to_bin(A().hostname))
            return
        if command != b'd':
            self.disconnect()
            lg.warn('not a "data" packet from %s : %r' % (str(self.transport.getPeer()), data))
            return
        inp = BytesIO(payload)
        try:
            file_id = int(struct.unpack('i', inp.read(4))[0])
            file_size = int(struct.unpack('i', inp.read(4))[0])
        except:
            inp.close()
            self.disconnect()
            lg.exc()
            lg.warn('wrong data from %s' % str(self.transport.getPeer()))
            return
        if self.fin is None:
            self.fin, self.fpath = tmpfile.make('idsrv', extension='.xml')
        inp_data = inp.read()
        inp.close()
        os.write(self.fin, inp_data)
        self.received += len(inp_data)
        self.sendString(version + b'o' + struct.pack('i', file_id))
        if self.received == file_size:
            os.close(self.fin)
            A('incoming-identity-file', self.fpath)
            self.fin = None
            self.fpath = None

    def connectionLost(self, reason):
        """
        """


#------------------------------------------------------------------------------


class IdServerFactory(ServerFactory):

    def buildProtocol(self, addr):
        p = IdServerProtocol()
        p.factory = self
        return p


#------------------------------------------------------------------------------


class WebMainPage(resource.Resource):

    def render_POST(self, request):
        inp = BytesIO(request.content.read())
        fin, fpath = tmpfile.make('idsrv', extension='.xml')
        inp_data = inp.read()
        inp.close()
        os.write(fin, inp_data)
        os.close(fin)
        A('incoming-identity-file', fpath)
        return ''

    def render_GET(self, request):
        src = ''

        src += '<div id="ui-id-servers" class="section bg-light">\n'
        src += '<div class="container">\n'
        src += '<div class="ui-card">\n'
        src += '<div class="card-body">\n'

        src += '<div class="row justify-content-center">\n'
        src += '<h1 align=center>Identities on %s</h1>' % strng.to_text(A().hostname)
        src += '</div><br>\n'

        src += '<div class="row">\n'
        src += '<table cellspacing=0 width=100% border=0><tr valign=top>\n'
        src += '<td width=152px nowrap>\n'
        HTDOCS_DIR = settings.IdentityServerDir()
        files = []
        if os.path.isdir(HTDOCS_DIR):
            for filename in os.listdir(HTDOCS_DIR):
                filepath = os.path.join(HTDOCS_DIR, filename)
                if os.path.isdir(filepath):
                    continue
                if not filename.endswith('.xml'):
                    continue
                files.append(filename)
        files.sort()
        currentChar = ''
        charIndex = 0
        for filename in files:
            if filename[0] != currentChar:
                currentChar = filename[0]
                if charIndex % 4 == 3:
                    src += '\n</td>\n<td width=152px nowrap>\n'
                charIndex += 1
                src += '\n<br>\n<h3>%s</h3>\n' % str(currentChar).upper()
            url = '/' + filename
            name = filename[:-4]
            src += '<p><a href="%s"><nobr>%s</nobr></a></p>\n' % (strng.to_text(url), strng.to_text(name))
        src += '</td>\n</tr>\n</table>\n'
        src += '<br><br><p>Total identities on "%s": %d</p><br><br>\n' % (strng.to_text(A().hostname), len(files))
        del files
        src += '<p>Other known identity servers:\n'
        for idhost in sorted(known_servers.by_host().keys()):
            idport = known_servers.by_host()[idhost][0]
            if idport != 80:
                idhost += b':%d' % idport
            src += '<a href="http://%s/"><nobr>%s</nobr></a>&nbsp;&nbsp;\n' % (strng.to_text(idhost), strng.to_text(idhost))
        src += '</p>'
        src += '<!--CLIENT_HOST=%s:%s-->\n' % (request.client.host, request.client.port)
        src += '</div>\n'
        src += '</div>\n'
        src += '</div>\n'
        src += '</div>\n'
        src += '</div>\n'

        html_src = web_html_template.WEB_ROOT_TEMPLATE % dict(
            title='Identities on %s' % strng.to_text(A().hostname),
            site_url='https://bitdust.io',
            basepath='https://identities.bitdust.io/',
            wikipath='https://bitdust.io/wiki/',
            idserverspath='https://identities.bitdust.io/',
            blockchainpath='https://blockchain.bitdust.io/',
            div_main_class='main idservers',
            div_main_body=src,
            google_analytics='',
            pre_footer='',
        )
        return strng.to_bin(html_src)


#------------------------------------------------------------------------------


class WebRoot(resource.Resource):

    def getChild(self, path, request):
        if not path:
            return self
        try:
            path = strng.to_text(path)
        except:
            return resource.NoResource('Not found')
        if not path.count('.xml'):
            return resource.NoResource('Not found')
        if path.count('.') != 1:
            return resource.NoResource('Not found')
        if path.count('/') or path.count('\\'):
            return resource.NoResource('Not found')
        filepath = os.path.join(settings.IdentityServerDir(), strng.to_text(path))
        if os.path.isfile(filepath):
            return static.File(filepath)
        return resource.NoResource('Not found')


#------------------------------------------------------------------------------


class KnownIDServersWebRoot(resource.Resource):

    def getChild(self, path, request):
        if not path:
            return self
        return resource.NoResource('Not found')


class KnownIDServersWebMainPage(resource.Resource):

    def render_GET(self, request):
        src = ''
        src += '<div id="ui-blockchain-explorer" class="section bg-light">\n'
        src += '<div class="container">\n'
        src += '<div class="ui-card">\n'
        src += '<div class="card-body">\n'

        src += '<div class="row justify-content-center">\n'
        src += '<h1 align=center>BitDust identity servers</h1>\n'
        src += '</div><br>\n'

        src += '<div class="row">\n'
        for idhost in sorted(known_servers.by_host().keys()):
            idport = known_servers.by_host()[idhost][0]
            if idport != 80:
                idhost += ':%d' % idport
            src += '<div class="col-4">\n'
            src += '<div class="ui-card ui-action-card ui-accordion shadow-sm" data-target="http://%s/">\n' % strng.to_text(idhost)
            src += '<div class="card-body">\n'
            src += '<h5 align=center><a class="card-link" href="http://%s/">%s</a></h5>\n' % (strng.to_text(idhost), strng.to_text(idhost))
            src += '</div>\n'
            src += '</div>\n'
            src += '</div>\n'
        src += '</div>\n'

        src += '</div>\n'
        src += '</div>\n'
        src += '</div>\n'
        src += '</div>\n'

        html_src = web_html_template.WEB_ROOT_TEMPLATE % dict(
            title='BitDust identity servers',
            site_url='https://bitdust.io',
            basepath='https://identities.bitdust.io/',
            wikipath='https://bitdust.io/wiki/',
            idserverspath='https://identities.bitdust.io/',
            blockchainpath='https://blockchain.bitdust.io/',
            div_main_class='main idservers',
            div_main_body=src,
            google_analytics='',
            pre_footer=web_html_template.pre_footer % dict(basepath='https://identities.bitdust.io/'),
        )
        return strng.to_bin(html_src)


def render_known_id_servers(web_port):
    root = KnownIDServersWebRoot()
    root.putChild(b'', KnownIDServersWebMainPage())
    reactor.listenTCP(web_port, server.Site(root))  # @UndefinedVariable


#------------------------------------------------------------------------------


def main():
    bpio.init()
    settings.init()
    lg.set_debug_level(_DebugLevel)

    serve_known_id_servers = False
    if len(sys.argv) > 1:
        if sys.argv[1].strip() == 'known_id_servers':
            serve_known_id_servers = True
            web_port = settings.getIdServerWebPort()
        else:
            web_port = int(sys.argv[1])
    else:
        web_port = settings.getIdServerWebPort()
    if len(sys.argv) > 2:
        tcp_port = int(sys.argv[2])
    else:
        tcp_port = settings.getIdServerTCPPort()

    if serve_known_id_servers:
        web_port = int(sys.argv[2])
        lg.out(2, 'serving known ID servers web_port=%d' % web_port)
        reactor.callWhenRunning(render_known_id_servers, web_port)  # @UndefinedVariable

    else:
        lg.out(2, 'starting ID server web_port=%d tcp_port=%d' % (
            web_port,
            tcp_port,
        ))
        reactor.addSystemEventTrigger('before', 'shutdown', A().automat, 'shutdown')  # @UndefinedVariable
        reactor.callWhenRunning(A, 'init', (web_port, tcp_port))  # @UndefinedVariable
        reactor.callLater(0, A, 'start')  # @UndefinedVariable

    reactor.run()  # @UndefinedVariable
    settings.shutdown()
    lg.out(2, 'reactor stopped, EXIT')


#------------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#!/usr/bin/python
# number.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (number.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
"""
.. module:: number.


"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
import sys
import struct

#------------------------------------------------------------------------------

if sys.version_info[0] == 2:

    def b(s):
        return s
else:

    def b(s):
        return s.encode('latin-1')  # utf-8 would cause some side-effects we don't want


#------------------------------------------------------------------------------


def long_to_bytes(n, blocksize=0):
    """long_to_bytes(n:long, blocksize:int) : string
    Convert a long integer to a byte string.

    If optional blocksize is given and greater than zero, pad the front of the
    byte string with binary zeros so that the length is a multiple of
    blocksize.
    """
    # after much testing, this algorithm was deemed to be the fastest
    s = b('')
    n = int(n)
    pack = struct.pack
    while n > 0:
        s = pack('>I', n & 0xffffffff) + s
        n = n >> 32
    # strip off leading zeros
    for i in range(len(s)):
        if s[i] != b('\000')[0]:
            break
    else:
        # only happens when n == 0
        s = b('\000')
        i = 0
    s = s[i:]
    # add back some pad bytes.  this could be done more efficiently w.r.t. the
    # de-padding being done above, but sigh...
    if blocksize > 0 and len(s) % blocksize:
        s = (blocksize - len(s) % blocksize)*b('\000') + s
    return s


def bytes_to_long(s):
    """bytes_to_long(string) : long
    Convert a byte string to a long integer.

    This is (essentially) the inverse of long_to_bytes().
    """
    acc = 0
    unpack = struct.unpack
    length = len(s)
    if length % 4:
        extra = (4 - length % 4)
        s = b('\000')*extra + s
        length = length + extra
    for i in range(0, length, 4):
        acc = (acc << 32) + unpack('>I', s[i:i + 4])[0]
    return acc
